
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "initng.h"

#include <time.h>                           /* time() */
#include <fcntl.h>                          /* fcntl() */
#include <sys/un.h>                         /* memmove() strcmp() */
#include <sys/wait.h>                       /* waitpid() sa */
#include <linux/kd.h>                       /* KDSIGACCEPT */
#include <sys/ioctl.h>                      /* ioctl() */
#include <stdio.h>                          /* printf() */
#include <stdlib.h>                         /* free() exit() */
#include <sys/reboot.h>                     /* reboot() RB_DISABLE_CAD */
#include <assert.h>
#include <errno.h>

#include "initng_global.h"

#include "initng_active_db.h"
#include "initng_toolbox.h"
#include "initng_main.h"
#include "initng_execute.h"
#include "initng_common.h"
#include "initng_load_module.h"

#include "initng_handler.h"
#include "initng_kill_handler.h"

#include "initng_static_data_id.h"

/* function defines */
static int try_to_start_service(active_h * service_to_start);
static int try_to_stop_service(active_h * service_to_stop);


    /* will try start processes waiting for dependencies to go up */
void try_start_wait_deps(void)
{
    active_h *current, *q = NULL;

    S_;

    /* walk through all active services */
    while_active_db_safe(current, q)
    {
        assert(current->name);
        /*D_("try_start_wait_deps(): re-trying with %s.\n", current->name); */

        /* Look at status */
        switch (current->a_status)
        {
            case WAITING_FOR_STOP_DEP:
                /* if succeeded to stop some, remember this */
                D_("Retry to stop service %s, that is WAITING_FOR_STOP_DEP.\n", current->name);
                try_to_stop_service(current);
                break;
            case WAITING_FOR_START_DEP:
                /* if succeeded to start some, remember this too */
                /* If start dependencies are suddenly met, but we're stopping the system try to stop */
                if (g.sys_state == STATE_STOPPING)
                {
                    F_("Service %s if WAITING_FOR_START_DEP, and will be stoped now!.\n", current->name);
                    try_to_stop_service(current);
                    break;
                }

                D_("Retry to start service %s, that is WAITING_FOR_START_DEP.\n", current->name);
                try_to_start_service(current);
                break;
            default:
                break;
        }
    }
}

/* this function will send term signal to a service if it was marked 30 seconds or more */
void term_stopping_services(void)
{
    active_h *to_term, *q = NULL;

    S_;

    while_active_db_safe(to_term, q)
    {
        if ((to_term->a_status == STOPPING ||
             to_term->a_status == STOP_LAUNCHING))
        {
            D_("Service %s is STOPPING\n", to_term->name);
            if (to_term->start_process && to_term->start_process->pid)
            {
                D_("Service %s has a start_process with a valid pid file\n",
                   to_term->name);
                if ((time(0) - to_term->time_got_status) >=
                    SECONDS_BEFORE_KILL)
                {
                    D_("Service %s has been STOPPING for %i, and its more then %i seconds\n", to_term->name, (time(0) - to_term->time_got_status), SECONDS_BEFORE_KILL);
                    /* todo, if kill(process,0) = dead then handle_killed_by_pid without warning and timeout. */

                    /* check if the process still exits! */
                    if (kill(to_term->start_process->pid, 0) < 0
                        && (errno == ESRCH))
                    {
                        /* This process is already dead */
                        F_("process %s pid %i is already dead, trying to fix.\n", to_term->name, to_term->start_process->pid);
                        handle_killed_by_pid(to_term->start_process->pid, 1);
                    }
                    else
                    {

                        W_("Warning, terminating %s on pid %i!\n",
                           to_term->name, to_term->start_process->pid);
                        /* Uhm, this doesnt work : kill(-to_term->start_process->pid, SIGKILL); */
                        kill(to_term->start_process->pid, SIGKILL);
                        /* sigchld might not trigger here, for example if we're using an invalid pidfile (pid = not ours?) */
                        /*                handle_killed_by_pid(to_term->start_process->pid, 1); */
                    }
                }
                else
                {
                    D_("Time is not out ( only %i seconds ) before start killing, waiting %i more seconds.\n", (time(0) - to_term->time_got_status), SECONDS_BEFORE_KILL);
                    alarm(SECONDS_BEFORE_KILL);
                }
            }
        }
    }
}

    /* M A R K    S T U F F */

    /* When this function is called, the service is marked for starting */
int start_service(active_h * service_to_start)
{
    assert(service_to_start);
    assert(service_to_start->name);

    S_;

    if (strncmp(service_to_start->name, "virtual/", 8) == 0)
    {
        F_("Cant start a virtual service, it has to be provided!\n");
        return (FALSE);
    }

    /* check system state, if we can launch. */
    if (g.sys_state != STATE_STARTING && g.sys_state != STATE_UP)
    {
        F_("Cant start service, when system status is: %i !\n", g.sys_state);
        return (FALSE);
    }

    /* check if its alredy starting */
    if (service_to_start->a_status == WAITING_FOR_START_DEP)
    {
        D_("service %s is starting allredy.\n", service_to_start->name);
        return (TRUE);
    }

    /* it might actually already be up */
    if (service_to_start->a_status == DONE ||
        service_to_start->a_status == RUNNING)
    {
        D_("service %s is alredy up!\n", service_to_start->name);
        return (TRUE);
    }

    /* if service has not started stopping yet, it can be set to up directly */
    if (service_to_start->a_status == WAITING_FOR_STOP_DEP ||
        service_to_start->a_status == STOP_DEP_MET)
    {
        /* set status to up or running again */
        if (service_to_start->done_on_exit == TRUE)
        {
            mark_service(service_to_start, DONE);
        }
        else if (service_to_start->up_when_running == TRUE)
        {
            mark_service(service_to_start, RUNNING);
        }
        else
        {
            F_("Setting this service to status RUNNING, this is a fallback.\n");
            mark_service(service_to_start, RUNNING);
        }

        D_("service %s did never have tome to stop!\n",
           service_to_start->name);
        return (TRUE);
    }

    /* STEP 2, service must be in one of these statuses to start */
    if (service_to_start->a_status != LOADING &&
        service_to_start->a_status != STOPPED &&
        service_to_start->a_status != RESPAWNING &&
        service_to_start->a_status != RESTARTING)
    {
        W_("cant set a service with status %i:%s, to start\n",
           service_to_start->a_status,
           active_db_get_status_string(service_to_start->a_status));
        return (FALSE);
    }

    /* make sure try_to_start_wait_deps is run on main loop */
    g.interrupt = TRUE;

    /* mark it WAITING_FOR_START_DEP and wait */
    if (!mark_service(service_to_start, WAITING_FOR_START_DEP))
    {
        W_("mark_service WAITING_FOR_START_DEP failed for service %s\n",
           service_to_start->name);
        return (FALSE);
    }
    return (TRUE);
}

    /* S T O P     S E R V I C E */
int stop_service(active_h * service_to_stop)
{
    assert(service_to_stop);

    D_("stop_service(%s);\n", service_to_stop->name);

    /* if service is alredy stoped, return */
    if (service_to_stop->a_status == STOPPED)
        return (TRUE);

    /* this can be a cirular dep, if then return */
    if (service_to_stop->a_status == STOP_MARKED ||
        service_to_stop->a_status == WAITING_FOR_STOP_DEP)
        return (TRUE);

    /* if service are waiting to start, there is no problem to shut it down directly */
    if (service_to_stop->a_status == WAITING_FOR_START_DEP ||
        service_to_stop->a_status == START_DEP_MET)
    {
        if (!mark_service(service_to_stop, STOP_DEP_MET))
            return (FALSE);
        return (TRUE);
    }

    /* check if service status, is really up */
    if (service_to_stop->a_status != DONE &&
        service_to_stop->a_status != RUNNING)
        W_("Warning, setting service %s, status %s, to STOP_MARKED!\n",
           service_to_stop->name,
           active_db_get_status_string(service_to_stop->a_status));

    /* set stopping */
    if (!mark_service(service_to_stop, STOP_MARKED))
        return (FALSE);

    /* stop all services depending on me */
    {
        active_h **deps = NULL;
        int i = 0;

        deps = active_db_dep_on_me(service_to_stop);
        while (deps[i])
        {
            D_("%s have to stop %s.\n", service_to_stop->name, deps[i]->name);
            stop_service(deps[i]);
            i++;
        }
        free(deps);
    }

    /* make sure, main loop tries to start WAITING_FOR_STARTDEP */
    g.interrupt = TRUE;

    /* Mark this service, that is is waiting for its stop deps to stop */
    if (!mark_service(service_to_stop, WAITING_FOR_STOP_DEP))
        return (FALSE);

    return (TRUE);
}

int restart_service(active_h * service_to_restart)
{
    S_;
    service_to_restart->m_rest = TRUE;
    return (stop_service(service_to_restart));
}



/* Load new process named */
active_h *load_new_service_named(const char *service)
{
    active_h *to_load = NULL;

    D_("  --  %s\n", service);

    assert(service);

    if (strncmp(service, "virtual/", 8) == 0)
    {
        W_("Virtuals can not be started manually, they have to be provided.\n");
        return (FALSE);
    }

    if (!(to_load = load_to_active(service)))
    {
        /* the function calling this function will print out an error */
        F_("Unable to load active for service %s\n", service);
        return (NULL);
    }

    /* STEP 3, mark service loading, if new */
    if ((to_load->a_status == NONE) && !mark_service(to_load, LOADING))
    {
        D_("Mark service failed!\n");
        return (NULL);
    }

    return (to_load);
}

    /* start new process named */
int start_new_service_named(const char *service)
{
    active_h *to_start = NULL;

    assert(service);
    D_("  --  %s\n", service);

    if (strncmp(service, "virtual/", 8) == 0)
    {
        W_("Virtuals can not be started manually, they have to be provided.\n");
        return (FALSE);
    }

    if (!(to_start = load_new_service_named(service)))
    {
        F_("Unable to load servie %s!\n", service);
        return (FALSE);
    }

    return (start_service(to_start));
}

    /* enter a new runlevel */
int stop_all(void)
{
    active_h *ser, *q = NULL;

    S_;
    set_sys_state(STATE_STOPPING);

    while_active_db_safe(ser, q)
    {
        stop_service(ser);
    }
    return (TRUE);
}





/* These two functions will stop services that are not dependencies 
   of the current runlevel */
static void mark_deps(active_h * h)
{
    W_("mark_deps(%s) Not implented!\n", h->name);
}

void stop_unneeded_services(void)
{
    active_h *s, *q = NULL;

    S_;
    /* TODO FIX THIS */
    W_("Not working TODO!\n");
    return;

    /* Remove mark from all active services */
    while_active_db(s)
    {
        s->m_needed = FALSE;
    }

    /* Recursively mark all dependencies of the current runlevel */
    if (!(s = active_db_find_by_name(g.runlevel)))
        return;
    mark_deps(s);

    /* Now rewalk the tree, stopping anything not marked as needed */
    s = NULL;
    while_active_db_safe(s, q)
    {
        if (s->m_needed == FALSE)
        {
            stop_service(s);
        }
    }
}



    /* D O    S T U F F */


    /* This is the function that really launches the start-marked services if all deps are up */
static int try_to_start_service(active_h * service_to_start)
{
    assert(service_to_start);

    D_("  --  (%s): status: \"%s\"\n",
       service_to_start->name,
       active_db_get_status_string(service_to_start->a_status));

    if (g.sys_state != STATE_STARTING && g.sys_state != STATE_UP)
    {
        F_("Cant start service, when system status is: %i !\n", g.sys_state);
        return (FALSE);
    }

    /* set status to START_DEP_MET */
    if (!mark_service(service_to_start, START_DEP_MET))
        return (FALSE);

    /* check if it has a start process */
    if (service_to_start->start_process)
    {
        F_("  --  (%s): Function already has a start process! Aborting\n",
           service_to_start->name);
        return (FALSE);
    }

    /* F I N A L L Y   S T A R T */

    switch (start_launch(service_to_start))
    {
        case FALSE:
            D_("Done direct!\n");
            /* D O N E - D I R E C T */
            if (!mark_service(service_to_start, DONE))
                return (FALSE);
            return (TRUE);                  /* return with status */
        case FAIL:
            F_("Did not find START file!\n");
            mark_service(service_to_start, FAIL_STARTING);
            return (FALSE);                 /* return with status */

    }

    /* make sure there are a start process */
    if (!service_to_start->start_process)
    {
        mark_service(service_to_start, FAIL_STARTING);
        return (FALSE);                     /* return with status */
    }

    D_("Launch Started.\n");

    if (service_to_start->done_on_exit == TRUE)
    {
        mark_service(service_to_start, STARTING);
    }
    else if (service_to_start->up_when_running == TRUE)
    {
        mark_service(service_to_start, RUNNING);
    }
    else
    {
        F_("Setting this service to status RUNNING, this is a fallback.\n");
        mark_service(service_to_start, RUNNING);
    }

    if (service_to_start->a_status == FAIL_STARTING)
        return (FALSE);
    return (TRUE);
}



/* do stop service */
static int try_to_stop_service(active_h * service_to_stop)
{
    int i = 0;
    active_h **deps = NULL;

    assert(service_to_stop);

    D_("  --  (%s): status :\"%s\"\n",
       service_to_stop->name,
       active_db_get_status_string(service_to_stop->a_status));

    /*
     *    Check so all deps, that needs service_to_stop, is down.
     *    if there are services depending on this one still running, return false and still try
     */
    deps = active_db_dep_on_me(service_to_stop);
    while (deps[i])
    {
        switch (deps[i]->a_status)
        {
            case DONE:
            case RUNNING:
            case STOP_MARKED:
            case STOPPING:
            case WAITING_FOR_STOP_DEP:
            case STOP_DEP_MET:
                D_("There is deps, depending on %s still, waiting a bit more\n", service_to_stop->name);
                free(deps);
                return (FALSE);             /* wait */
            default:
                break;
        }
        i++;
    }
    free(deps);

    /* ok, stopping deps are met */
    if (!mark_service(service_to_stop, STOP_DEP_MET))
    {
        D_("Cound not set status of %s to STOP_DEP_MET.\n",
           service_to_stop->name);
        return (FALSE);
    }

    /* check if it has a stop process */
    if (service_to_stop->stop_process)
    {
        F_("  --  (%s): Function already have a stop process! Aborting!\n",
           service_to_stop->name);
        mark_service(service_to_stop, FAIL_STOPPING);
        return (FALSE);
    }

    /* F I N A L L Y   S T O P */
    if (!mark_service(service_to_stop, STOPPING))
    {
        D_("Cound not set status of %s to STOPPING.\n",
           service_to_stop->name);
        return (FALSE);
    }

    /* launch stop service */
    if (stop_launch(service_to_stop) == FAIL)
    {
        F_("  --  (%s): fail launch stop!\n", service_to_stop->name);
        mark_service(service_to_stop, FAIL_STOPPING);
        return (FALSE);
    }


    /* if we got a stop process with a pid */
    if (service_to_stop->stop_process && service_to_stop->stop_process->pid)
    {
        alarm(SECONDS_BEFORE_KILL + 1);     /* the alarm will try to kill stopping services */
        D_("Clock is set, %s stop script, have %i seconds to kill the service.\n", service_to_stop->name, SECONDS_BEFORE_KILL);
        return (TRUE);                      /* return from here */
    }

    /* T R Y 2 , Kill start_process directly */
    if (service_to_stop->start_process
        && service_to_stop->start_process->pid > 0)
    {
        D_("  --  (%s): REAL KILLING PID %i!\n",
           service_to_stop->name, service_to_stop->start_process->pid);

        /* Uhm, this doesnt work : kill(-service_to_stop->start_process->pid, SIGTERM); */
        kill(service_to_stop->start_process->pid, SIGTERM);

        alarm(SECONDS_BEFORE_KILL + 1);     /* the alarm will try to kill stopping services */
        D_("Clock is set, killed process %i (%s), have %i seconds to quit before SIGKILL\n", service_to_stop->start_process->pid, service_to_stop->name, SECONDS_BEFORE_KILL);
        /*kill(service_to_stop->start_process->pid,SIGKILL); */
        return (TRUE);
    }
    /* if this is a empty dep service we show up here */

    /* remember this in history */
    if (!mark_service(service_to_stop, STOPPED))
    {
        D_("Coud not set status STOPPED!\n");
        return (FALSE);
    }

    /* check if this service is restarting */
    if (service_to_stop->m_rest == TRUE)
    {
        if (!mark_service(service_to_stop, RESTARTING))
            return (FALSE);
        service_to_stop->m_rest = FALSE;
        start_service(service_to_stop);
        D_("Service is restarting now!\n");
        return (TRUE);
    }


    /* free service, and forget */
    active_db_del(service_to_stop);
    active_db_free(service_to_stop);
    D_("Service removed.\n");
    return (TRUE);
}
