
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "initng.h"

#include <time.h>                           /* time() */
#include <fcntl.h>                          /* fcntl() */
#include <sys/un.h>                         /* memmove() strcmp() */
#include <sys/wait.h>                       /* waitpid() sa */
#include <linux/kd.h>                       /* KDSIGACCEPT */
#include <sys/ioctl.h>                      /* ioctl() */
#include <stdio.h>                          /* printf() */
#include <stdlib.h>                         /* free() exit() */
#include <sys/reboot.h>                     /* reboot() RB_DISABLE_CAD */
#include <assert.h>

#include "initng_global.h"
#include "initng_handler.h"
#include "initng_history.h"
#include "initng_process_db.h"
#include "initng_execute.h"
#include "initng_toolbox.h"
#include "initng_main.h"
#include "initng_common.h"

#include "main.h"
#include "initng_service_cache.h"
#include "initng_active_db.h"
#include "initng_load_module.h"
#include "initng_plugin_callers.h"
#include "initng_fd.h"

#include "initng_kill_handler.h"
#include "initng_static_data_id.h"

static void handle_killed_start(active_h * killed_service);
static void handle_killed_stop(active_h * killed_service);

/* called when a process got killed, identify it, and make a call with a pointer to the process */
void handle_killed_by_pid(pid_t kpid, int r_code)
{
    /* The process that got killed */
    active_h *killd_p = NULL;

    D_("handle_killed_by_pid(%i);\n", kpid);

    /* don't do anything on this pid */
    if (kpid <= 1)
        return;

    /* Look in process database for a match */
    if (!(killd_p = active_db_find_by_pid(kpid)))
    {
        D_("handle_killed_by_pid(%i): No match!\n", kpid);
        return;
    }

    D_("handle_killed_by_pid(%i): found service \"%s\"...\n", kpid,
       killd_p->name);

    if (killd_p->start_process && killd_p->start_process->pid == kpid)
    {
        D_("handle_killed_by_pid(%i): seems to be a start process killed...\n", kpid);
        killd_p->start_process->r_code = r_code;
        handle_killed_start(killd_p);
        return;
    }
    else if (killd_p->stop_process && killd_p->stop_process->pid == kpid)
    {
        D_("handle_killed_by_pid(%i): seems to be a stop process killed...\n",
           kpid);
        killd_p->stop_process->r_code = r_code;
        handle_killed_stop(killd_p);
        return;
    }

    F_("handle_killed_by_pid(%i): traced killed pid to a service but can't find servicetype!\n", kpid);
    return;
}

/*
 * handle_killed_start(), description:
 *
 * A process dies, and kill_handler walks the active_db looking for a match
 * on that pid, if it founds a match for a active->start_process, this
 * function is called, with a pointer to that active_service to
 * handle the situation.
 */
static void handle_killed_start(active_h * killed_service)
{
    assert(killed_service);
    assert(killed_service->name);

    D_("handle_killed_start(%s): initial status: \"%s\".\n",
       killed_service->name,
       active_db_get_status_string(killed_service->a_status));

    /* Set the universal variable, that signalize that something happend */
    g.interrupt = TRUE;

    /* Notify readpipe plugins that we're dead and pipes will be closed */
    initng_process_readpipe(killed_service, killed_service->start_process,
                            TRUE);

    /* Check if a plugin wants to override handle_killed behavior */
    if (initng_plugin_handle_killed
        (killed_service, killed_service->start_process))
    {
        D_("Plugin did handle this kill.\n");
        return;
    }

    /*
     * If the return code (for example "exit 1", in a bash script)
     * from the program, is bigger than 0, this commonly signalize
     * that something got wrong, print this as an error msg on screen
     */
    if (killed_service->start_process->r_code > 0)
        F_("handle_killed_start(%s): Returned with exit %i.\n",
           killed_service->name, killed_service->start_process->r_code);

    /*
     * If there exists an stop_process, it probably that process that did kill
     * the start_process, that may be an daemon. If its so, do nothing that is
     * probably done in handle_killed_stop();
     */
    if (killed_service->stop_process)
    {
        /* TODO, shud we not free the start_process here, where is that done anyway? */
        W_("Start process stopped, now waiting for stop process to stop!\n");
        return;
    }

    /*
     * If this active_service status is STOPPING, the exit of this daemon, is probably
     * A Good Thing (tm), and wont need special handeling.
     */
    if (killed_service->a_status == STOPPING ||
        killed_service->a_status == STOP_LAUNCHING ||
        killed_service->a_status == STOP_MARKED)
    {
        mark_service(killed_service, STOPPED);
        D_("a_status == aka STOPPING, mark STOPPED and delete from active_db\n");

        /* gotos are not nice, this walks forward in the function, to the stopped service part */
        goto stop_this;
    }

    /*
     * This active_service might be a service, that is designed to be up and started,
     * when the code is run, and then stoped, detect this, and set service to status
     * DONE to mark this
     */
    if (killed_service->done_on_exit == TRUE)
    {
        D_("This service, is up when it dies, setting it to up now...\n");

        /* if return code, is positive, this probably is a
         * bad thing (override with option UP_WHEN_FAIL),
         * mark this process FAIL_STARTING, so user notice this.
         */
        if (!killed_service->start_process->r_code
            || active_db_is(&UP_WHEN_FAIL, killed_service))
            mark_service(killed_service, DONE);
        else
            mark_service(killed_service, FAIL_STARTING);

        /* free the process struct, to spare memory */
        process_db_free(killed_service->start_process);
        killed_service->start_process = NULL;
        return;

    }

    /* OK, now accept this service as STOPPED, set this status */
    mark_service(killed_service, STOPPED);


    /*
     * if this service is set to respawn, (By the RESPAWN option)
     * do it, but don't respawn services on system state
     * STATE_STOPPING (system rebooting, or halting) or if
     * service is marked restarting.
     */
    if ((active_db_is(&RESPAWN, killed_service))
        && (g.sys_state != STATE_STOPPING))
    {
        D_("Respawn is set, now respawning\n");
        mark_service(killed_service, RESPAWNING);
        /* free the start_process, or start_service will fail */
        process_db_free(killed_service->start_process);
        killed_service->start_process = NULL;
        /* respawn */
        start_service(killed_service);
        return;
    }

    /*
     * FAQ:
     * A: What differs RESPAWN != RESTART?
     * Q: RESPAWN: the process is restarted, becouse it died,
     *    on RESTART: the process is restarted, on a command from
     *    the user, thru ngc -r service.
     */

    /*
     * If this stop, is planed restart and requested by the user
     * restart the service now.
     */
    if (killed_service->m_rest == TRUE)
    {
        D_("Restarting set, now restarting\n");
        mark_service(killed_service, RESTARTING);
        killed_service->m_rest = FALSE;
        /* free the start_process, or start_service will fail */
        process_db_free(killed_service->start_process);
        killed_service->start_process = NULL;
        /* respawn */
        start_service(killed_service);
        return;
    }
    /* goto stop_his */

    /* i hate gotos, but this is good avoiding duplicating code */
  stop_this:

    /* OK, now aceppt this service as stoping, no HOKUS POKUS */
    D_("Okay, service %s is really stopping\n", killed_service->name);

    /*
     * This service where killed, and therefor, we need to find all services
     * that depends on this werry service that was killed, and stop that to,
     * if running
     */
    {
        int i = 0;
        active_h **deps = active_db_dep_on_me(killed_service);

        /* mark all services depending on this service stop */
        while (deps[i])
        {
            D_("%s have to stop %s.\n", killed_service->name, deps[i]->name);
            stop_service(deps[i]);
            i++;
        }
        free(deps);
    }

    D_("Now removing service %s\n", killed_service->name);


    /*
     * This removes service from active_list, and free(), the data.
     */
    active_db_del(killed_service);
    active_db_free(killed_service);
}


/* to do when a stop action dies */
static void handle_killed_stop(active_h * killed_service)
{
    assert(killed_service);

    g.interrupt = TRUE;
    D_("handle_killed_stop(%s);\n", killed_service->name);

    /* Notify readpipe plugins that we're dead and pipes will be closed */
    initng_process_readpipe(killed_service, killed_service->stop_process,
                            TRUE);

    /* Check if a plugin wants to override handle_killed behavior */
    if (initng_plugin_handle_killed
        (killed_service, killed_service->stop_process))
        return;

    /* free stopped process data */
    process_db_free(killed_service->stop_process);
    killed_service->stop_process = NULL;


    /* mark service stopped */
    mark_service(killed_service, STOPPED);

    if (killed_service->m_rest == TRUE)
    {
        mark_service(killed_service, RESTARTING);
        killed_service->m_rest = FALSE;

        /* also free start_process */
        if (killed_service->start_process)
        {
            process_db_free(killed_service->start_process);
            killed_service->start_process = NULL;
        }

        /* respawn */
        start_service(killed_service);
        return;
    }


    /* if there is no start process, count this as the end of service */
    if (killed_service->start_process)
    {
        F_("Bad, there is a start process!\n");
        return;
    }

    /* free service, if stopped = disappear from active list */
    active_db_del(killed_service);
    active_db_free(killed_service);
}
